#!/usr/bin/env ruby

class Gui_support

attr_reader :update_ripping_progress, :update_encoding_progress, :append_2_log, :track_finished, :finished, :error_message, :ripping_progress, :encoding_progress, :mismatch

	def initialize(gui, dirnames) #gui is an instance of the graphical user interface used
		@gui = gui
		@logfiles = Array.new
		dirnames.each{|dir| @logfiles << File.open(dir + "/ripping.log", "a")} # Open a log file in append mode for each directory
		@problem_tracks = Hash.new # key = tracknumber, value = new dictionary with key = seconds_chunk, value = [amount_of_chunks, trials_needed]
		@not_corrected_tracks = Array.new # Array of tracks that weren't corrected within the maximum amount of trials set by the user
		@ripping_progress = false
		@encoding_progress = false
		append_2_log("This log is created by Rubyripper, version 0.3\n\n", "initialize")
		append_2_log("Official website: http://rubyforge.org/projects/rubyripper/\n", "initialize")
		append_2_log("Wiki page: http://wiki.hydrogenaudio.org/index.php?title=Rubyripper\n\n", "initialize")
	end
	
	def update_ripping_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @ripping_progress = new_value : @ripping_progress = 1.0
		@gui.update("ripping_progress", @ripping_progress)
	end
	
	def update_encoding_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @encoding_progress = new_value : @encoding_progress = 1.0
		@gui.update("encoding_progress", @encoding_progress)
	end
	
	def append_2_log(message, calling_function = false)
		@logfiles.each{|logfile| logfile.print(message); logfile.flush()} # Append the messages to the logfiles
		@gui.update("log_change", message)
	end
	
	def update_logfiles(message)
		@logfiles.each{|logfile| logfile.print(message)} # Append the messages to the logfiles
	end
	
	def track_finished(track, calling_function = false) #Function needed to unselect tracks that are finished
		@gui.update("track_finished", track)
		# When first track is ripped, hint the gui that encoding is started
		unless @encoding_progress : @encoding_progress = 0.0 ; update_encoding_progress(0.0, "track_finished") end 
	end
	
	def finished(matches_all, matches_errors, maxtries, calling_function = false) #Function needed to lock the gui for accepting new cd's.
		summary(matches_all, matches_errors, maxtries)
		@gui.update("finished", true)
	end
	
	def error_message(message, calling_function = false)
		@gui.update("error", message)
	end
	
	def mismatch(track, trial, indexes_with_errors, size, length)
		if !@problem_tracks.key?(track) #First time we encounter this track (Secure_rip->analyze_files() )
			@problem_tracks[track] = Hash.new # create the Hash for the track
			indexes_with_errors.each do |index_of_chunk|
				seconds = ((index_of_chunk.to_f / size) * length).to_i # position of chunk rounded in seconds
				if !@problem_tracks[track].key?(seconds)
					@problem_tracks[track][seconds] = [1, trial] # different_chunks, trial. First time we encounter this position, so different_chunks = 1
				else
					@problem_tracks[track][seconds][0] += 1 # one more chunk at the same second
				end
			end
		else
			indexes_with_errors.each do |index_of_chunk|
				seconds = ((index_of_chunk.to_f / size) * length).to_i # position of chunk rounded in seconds
				@problem_tracks[track][seconds][1] = trial #Update the amount of trials needed
			end
		end
		if trial == 0: @not_corrected_tracks << track end #Reached maxtries and still got errors
	end
	
	def summary(matches_all, matches_errors, maxtries) #Give an overview of errors
		append_2_log("Rubyripper is finished with the ripping and encoding of selected tracks.\n")
		append_2_log("\nRIPPING SUMMARY\n\n")
		
		append_2_log("All chunks were tried to match at least #{matches_all} times.\n")
		if matches_all != matches_errors: append_2_log("Chunks that differed after #{matches_all} trials, were tried to match #{matches_errors} times.\n") end

		if @problem_tracks.empty?
 			append_2_log("None of the tracks gave any problems\n")
 		elsif @not_corrected_tracks.size != 0
 			append_2_log("Some track(s) could NOT be corrected within the maximum amount of trials\n")
 			@not_corrected_tracks.each do |track|
				append_2_log("Track #{track} could NOT be corrected completely\n")
			end
 		else
 			append_2_log("Some track(s) needed correction, but could be corrected within the maximum amount of trials\n")
 		end

 		if !@problem_tracks.empty? # At least some correction was necessary
 			@gui.update("log_change", "Analyzing the position of chunks that needed correction...\n")
			position_analyse(matches_errors, maxtries)
			@gui.update("log_change", "The exact positions of the suspicious chunks can be found in the ripping log\n")
		end
 	end
 		
 	def position_analyse(matches_errors, maxtries) # Give an overview of suspicion position in the logfile
		update_logfiles("\nSUSPICION POSITION ANALYSIS\n\n")
		update_logfiles("Since there are more than 150 chunks per second, after making the notion of the\n")
		update_logfiles("suspicion position, the amount of initially mismatched chunks for that position is shown.\n\n")
		@problem_tracks.keys.sort.each do |track| # For each track show the position of the files, how many chunks of that position and amount of trials needed to solve
			update_logfiles("TRACK #{track}\n")
			@problem_tracks[track].keys.sort.each do |length| #length = total seconds of suspicious position
				minutes = length / 60 # ruby math -> 70 / 60 = 1 (how many times does 60 fit in 70)
				seconds = length % 60 # ruby math -> 70 % 60 = 10 (leftover)
				if @problem_tracks[track][length][1] != 0
					update_logfiles("\tSuspicion position : #{sprintf("%02d", minutes)}:#{sprintf("%02d", seconds)} (#{@problem_tracks[track][length][0]}x) (CORRECTED at trial #{@problem_tracks[track][length][1] + 1})\n")
				else # Position could not be corrected
					update_logfiles("\tSuspicion position : #{sprintf("%02d", minutes)}:#{sprintf("%02d", seconds)} (#{@problem_tracks[track][length][0]}x) (COULD NOT BE CORRECTED)\n")
				end
			end
		end
		@logfiles.each{|logfile| logfile.close} #Close each file
	end
end

class Cuesheet # Started the very beginning
	def initialize(cddb, output_dirs, file)
		@cddb = cddb
		@output_dirs = output_dirs
		@cuesheet = ''
		@file = file
		toc = `cdparanoia #{@cdrom} 2>&1`.split("\n") #command
		
	end
	
	def analyze_toc # fill startpos, endpos, pregap
		@startpos = Array.new #Where does each track start?
		@endpos = Array.new #Where does each track end?
		@pregap = Array.new #Any space between tracks?
		
		while toc[0][0,2] != "==" : toc.shift end #remove all the non-interesting stuff
		toc.shift() #remove the ============ line
			
		@cddb.audiotracks.times do |time|
			temp = toc[time].split() # temp = [track, lengt_sector, length_time, start_sector]
			@startpos << temp[3].to_i # get the startpos
			@endpos << temp[1].to_i + temp[3].to_i #endpos = length + startpos
		end
		
		@startpos.size.times do |time| # Save pregap
			if time == 0: @pregap << @startpos[time]
			else @pregap << @startpos[time] - @endpos[time - 1]
			end
		end
	end
	
	def time(sector) # minutes:seconds:leftover frames
		minutes = sector / 4500 # 75 frames/second * 60 seconds/minute
		seconds = (sector % 4500) / 75
		frames = sector % 75 # leftover
		return "#{sprintf("%02d", minutes)}:#{sprintf("%02d", seconds)}:#{sprintf("%02d", frames)}"
	end
	
	def cuesheet_singlefile()
		@cuesheet += "REM GENRE #{@cddb.genre}\n"
		@cuesheet += "REM DATE #{@cddb.year}\n"
		@cuesheet += "COMMENT \"Rubyripper 0.3\"\n"
		@cuesheet += "PERFORMER \"#{@cddb.artist}\"\n"
		@cuesheet += "TITLE \"#{@cddb.album}\"\n"
		@cuesheet += "FILE \"#{@file}\" WAVE\n" #@file = filename
		@cddb.audiotracks.times do |track|
			@cuesheet += "\tTRACK #{sprintf("%02d", track + 1)} AUDIO\n"
			@cuesheet += "\t\tTITLE \"#{@cddb.tracklist[track]}\"\n"
			@cuesheet += "\t\tPERFORMER \"#{@cddb.artist}\"\n"
			if track > 0 && @pregap[track] > 0: @cuesheet += "\t\tINDEX 00 #{time(@endpos[track - 1])}\n" #Only if there is a pregap
			elsif track == 0 && @pregap[0] > 0: @cuesheet += "\t\tINDEX 00 00:00:00\n"
			end
			@cuesheet += "\t\tINDEX 01 #{time(@startpos[track])}\n"
		end
	end
end

class Cddb

attr_reader :audiotracks, :raw_response, :tracks_seconds, :make_internet_contact, :cdrom, :devicename, :totalsectors, :startsectors, :length_in_sectors, :tracks_playtime, :cd_playtime, :freedb, :freedb_choice
attr_accessor :artist, :album, :genre, :year, :tracklist
	
	def initialize(cdrom, instance, verbose=false)
		@cdrom = cdrom
		@gui = instance
		@verbose = verbose
		
		if !File.exists?(`which cdparanoia`.strip()) # Check for cdparanoia
			@gui.update("error", "cdparanoia is missing on your system.\nCan't analyze CD!")
		else 
			analyze_cd()
			set_default_values()
		end
	end
	
	def analyze_cd # based on cdparanoia output
		@audiotracks = 0
		@totalsectors = 0
		@devicename = "Unknown"
		@startsectors = Array.new
		@length_in_sectors = Array.new
		@tracks_playtime = Array.new
		@tracks_seconds = Array.new
		@cd_playtime = '00:00'
		output = `cdparanoia #{@cdrom} -vQ 2>&1`.split("\n") #command
		
		output.each do |line|
 			if line[0,5] =~ /\s+\d+\./ #Bingo! We've got the info of a track s = space, d = digit, + means 1 or more
				trackinfo = line.split() # Split the values tracknumber, length in sectors, playtime, startsector, starttime
				@audiotracks += 1 # each line present one track
				@startsectors << trackinfo[3].to_i #save the integer value
				@length_in_sectors << trackinfo[1].to_i
				@tracks_playtime << trackinfo[2][1,5]
			elsif line =~ /CDROM\D*:/
				@devicename = $'.strip()
			elsif line =~ /Unable to open disc/
				@gui.update("error", "No audio disc found in #{cdrom}\n")
				break
			elsif line[0,5] == "TOTAL"
				@totalsectors = line.split()[1].to_i
				@cd_playtime = line.split()[2][1,5]
			end
		end
		@length_in_sectors.each{|sectors| @tracks_seconds << sectors / 75} # Each second has 75 sectors
	end
	
	def set_default_values # for file naming and metadata
		@artist = @album = @genre = 'Unknown'; @year = '0'
		@tracklist = Array.new # Titles of the tracks
		@audiotracks.times{|number| @tracklist << "Track #{number + 1}"} # temporary tracklist
	end
	
	def freedb(freedb_options=false, choose=false) # based on cd-discid output
		@freedb_options = freedb_options ? freedb_options : {"site" => "freedb.org", "username" => "anonymous", "hostname" => "my_secret.com"} # if not set, use defaults
		@choose = choose # pass multiple freedb hits to gui
	
		if !File.exists?(`which cd-discid`.strip()) : @gui.update("error", "cd-discid is missing on your system.\nCan't fetch freedb info!") # Check if cd-discid is installed
 		else @freedb_string = `cd-discid #{@cdrom}`
 			if @freedb_string != "" # audio-disc is found
 				make_internet_contact()
 			else
 				@gui.update("error", "No audio disc found in #{cdrom}\n")
 			end
 		end
	end
	
	def make_internet_contact
		require 'socket'
		begin
			@socket = TCPSocket.new(@freedb_options['site'], 8880)
			@clientname = 'rubyripper'
			@version = '0.3'
		rescue
			@gui.update("error", "Couldn't connect to freedb server.\nAre you sure you're connected?\n")
		else
			handshake()
			cddb_query()
		end
	end	
		
	def handshake # make first contact and set protocol to 6
		@socket.gets # Get general welcome message
		@socket.puts("cddb hello #{@freedb_options["username"]} #{@freedb_options["hostname"]} #{@clientname} #{@version}") #Identify ourselves
		@socket.gets # Get specific welcome message
		@socket.puts('proto 6') # Set protocol to the latest & greatest
		@socket.gets # Get confirmation
	end
	
	def cddb_query # ask for matches on cd, if there are multiple, interaction with user is possible
		@socket.puts "cddb query #{@freedb_string}" # Fire query
		answer = @socket.gets # The first 3 numbers in the answer indicates if none, one or multiple hits are found
		if answer[0..2] == '200'  #There was only one hit found
			temp, @category, @discid = answer.split()
			freedb_choice(0, true)
		elsif answer[0..2] == '211' || answer[0..2] == '210' #Multiple hits were found
			get_choices()
			if !@choose : freedb_choice(0) #Always choose the first one
			else @gui.update("cddb_hit", @choices) #Let the user choose
			end	
		elsif answer[0..2] == '202'
			@gui.update("error", "Cddb couldn't find a match for the audio-disc.\nDefault values are used.\n")
		else
			@gui.update("error", "Return code = #{answer[0..2]}.\nReturn code is not supported yet.\n")
		end
	end
	
	def freedb_choice(choice, single=false) #if single=true there's no choise, otherwhise the choice is used for raw_response
		unless single
			@category, @discid = @choices[choice].split
		end
		get_raw_response()
		@tracklist.clear() #Now fill it with the real tracknames
		cd_vars()
		@gui.update("cddb_hit", false) # Give the signal that we're finished
	end

	def get_choices	#Puts the multiple choices in an array
		@choices = Array.new
		until (answer = @socket.gets.strip()) == '.' # This means end of output
			@choices << answer
		end
	end

	def get_raw_response #Retrieve all usefull metadata into @raw_response
		@socket.puts("cddb read #{@category} #{@discid}") # Get match
		answer = @socket.gets # Announcement of match
		@raw_response = Array.new
		until (answer = @socket.gets.strip()) == '.' # This means end of output
			if answer != nil && answer[0,1] != '#' && answer[-1,1] != '=' # Skip the garbage
 				@raw_response << answer
			end
		end
		@socket.puts("quit") # Say goodbye to server
		@socket.gets # Confirmation
		@socket.close
	end
	
	def cd_vars #Make some usefull variables from the raw_response.
		@raw_response.each do |line|
			if line =~ /DTITLE=/
				(@artist, @album) = $'.split(/\s*\/\s*/) # Remove the '/' with spaces around it
				if @artist != nil: clean(@artist) end
				if @album != nil: clean(@album) end
			elsif line =~ /DYEAR=/
				@year = $'
			elsif line =~ /DGENRE=/
				@genre = $'
			elsif line =~ /TTITLE\d*=/
				trackname = $'
				if trackname != nil: @tracklist << clean(trackname) end
			end
		end
	end
	
	def clean(var)
		var = var.strip
		if var.index(/[\[\]\/]/) # Don't accept names with a '\' or '[' or ']' in it. Otherwise neglect that part of the string.
			var = $`
		end
		var = var.rstrip
		return var
	end	
end

class Secure_rip
	def initialize(track, settings)
		@track = track
		@settings = settings
		@gui = @settings['log']
		@req_matches_all = @settings['req_matches_all'] # Matches needed for all chunks
		@req_matches_errors = @settings['req_matches_errors'] # Matches needed for chunks that didn't match immediately
		if @settings['offset'] != 0 && @track == @settings['cd'].audiotracks && @settings['rippersettings'].include?('-Z') #workaround for bug in cdparanoia
			@settings['rippersettings'].gsub!(/-Z\s?/, '') #replace the -Z setting (and one space if it is there) with nothing. See issue nr. 13.
		end
		main()
#eject the disc when finished if user checked this option
		if @settings['eject'] && @track == @settings['tracks_to_rip'][-1] && File.exists?(`which eject`.strip) : `eject #{@settings['cd'].cdrom}` end
	end
	
	def main
		@errors = Hash.new()
		@filesizes = Array.new
		@trial = 0
		
		@req_matches_all.times do # The amount of matches required tells how many times the files should be ripped before comparing
			@trial += 1
			rip()
			@filesizes << File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav")
		end
		
		test_for_filesize() #Make sure that we got the needed files with the same size
		analyze_files() #If there are differences, save them in the @errors hash
				
		while @errors.size > 0
			if @trial > @settings['max_tries'] && @settings['max_tries'] != 0 # We would like to respect our users settings, wouldn't we?
				@gui.append_2_log("Maximum tries reached. #{@errors.length} chunk(s) didn't match the required #{@req_matches_errors} times\n")
				@gui.append_2_log("Will continue with the file we've got so far\n")
				@gui.mismatch(@track, 0, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track - 1]) # zero means it is never solved.
				break # break out loop and continue using trial1
			end
			
			@trial += 1
			rip()
			
			if File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") == @size_match # The file size must match with previous attempts
				if @trial > @req_matches_errors : wave_out() else read_positions_only() end # If the req_matches errors is equal of higher to @trial, no match would ever be found, so skip
			else
				@gui.append_2_log("File size of trial #{@trial} does not match previous attempts\n")
			end
		end
		
		get_digest() # Get a MD5-digest for the logfile
		
		@gui.update_ripping_progress(@settings['rip_percentages'].shift)
	end
	
	def test_for_filesize
		@size_match = false # Until proven otherwise :P
		if @filesizes.uniq.size == 1 : @size_match = @filesizes[0] end # When there is no difference in the filesize proceed (this should happen in 99% of all cases)
		while !@size_match
			@gui.append_2_log("Still not #{@req_matches_all} times the same filesize. Trying another time\n")
			@trial += 1 ; rip() ; @filesizes << File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav")
			filesizes = @filesizes.sort() # Make a sorted copy
			filesizes.each do |size| # This look like a lot of iteration, but most of the code is only run once
				if filesizes.rindex(size) - filesizes.index(size) >= (@req_matches_all - 1) # Do we have the required amount of same sizes?
					@size_match = size
					trialnumber = 1 # So we have to iterate again with trialnumber as a help value.
					@filesizes.each_index do |index| # now clean up files with wrong filesizes. Since we sorted the copy, we don't know which trials have wrong filesizes.
						if @filesizes[index] == @size_match 
							if File.exist?("#{@settings['temp_dir']}track#{@track}_#{trialnumber}.wav") : trialnumber += 1 # The file won't exist if it is deleted earlier in the loop
							else File.rename("#{@settings['temp_dir']}track#{@track}_#{index + 1}.wav", "#{@settings['temp_dir']}track#{@track}_#{trialnumber}.wav") ; trialnumber += 1
							end	
  						else
  							File.delete("#{@settings['temp_dir']}track#{@track}_#{index + 1}.wav") # Delete files with non-matching filesizes
  						end
  					end
  					break # Since we have a filesize match, don't finish the loop
				end
			end
		end
	end
	
	def analyze_files()
		@gui.append_2_log("Analyzing files for mismatching chunks\n")
		files = Array.new
		@req_matches_all.times do |time|
			files << File.new("#{@settings['temp_dir']}track#{@track}_#{time + 1}.wav", 'r')
		end
				
		(@req_matches_all - 1).times do |time|
			index = 0 ; files.each{|file| file.pos = index}
			while index < @size_match
				if !@errors.key?(index) && files[0].read(1000) != files[time + 1].read(1000) # Does this chunk matches the previous ones? and isn't the position already known?
					files.each{|file| file.pos = index} # Reset each read position of the files
					@errors[index] = Array.new
					files.each{|file| @errors[index] << file.read(1000)} # Save the chunk for all files in the just created array
				end
				index += 1000
			end
		end
		files.each{|file| file.close}
		(@req_matches_all - 1).times{|time| File.delete("#{@settings['temp_dir']}track#{@track}_#{time + 2}.wav")} # Don't need these files anymore. Differences are saved in memory.
		if @errors.size == 0
			@gui.append_2_log("Every chunk matched #{@req_matches_all} times :)\n")
		else
			@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
			@gui.append_2_log("#{@errors.length} chunk(s) didn't match #{@req_matches_all} times.\n")
		end
	end
	
	def read_positions_only # This is used if the required amount of matches is too big to ever be realized
		file = File.new("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk|
			file.pos = start_chunk
			@errors[start_chunk] << file.read(1000)
		end
		file.close
		File.delete("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
	end
	
	def wave_out
	#Let the errors 'wave' out. Chunks with different outcomes must be matches @req_matches_errors times.
		file1 = File.new("#{@settings['temp_dir']}track#{@track}_1.wav", 'r+')
		file2 = File.new("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk| # Sort the hash keys to prevent jumping forward and backwards in the file
			file2.pos = start_chunk
			@errors[start_chunk] << temp = file2.read(1000)
			@errors[start_chunk].sort! # Sort the array before looking if temp solves our requirement of matches
			if (@errors[start_chunk].rindex(temp) - @errors[start_chunk].index(temp)) == (@req_matches_errors - 1) # Matches are r(ight)index - index because array is sorted
				file1.pos = start_chunk
				file1.write(temp)
				@errors.delete(start_chunk)
			end
		end
		file1.close
		file2.close
		File.delete("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		
		if @errors.size == 0
			@gui.append_2_log("Error(s) succesfully corrected, #{@req_matches_errors} matches found for each chunk :)\n")
		else
			@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
			@gui.append_2_log("#{@errors.length} chunk(s) didn't match #{@req_matches_errors} times.\n")
		end
	end
	
	def rip # set cdparanoia command + parameters
		@gui.append_2_log("Starting to rip track #{@track}, trial #{@trial}#\n")
		command = "cdparanoia"
		if @settings['rippersettings'].size != 0 : command += " #{@settings['rippersettings']}" end 
		command += (" #{@track}") # second parameter tracknumber
		command += " -d #{@settings['cdrom']}"
		command += " -O #{@settings['offset']}"
		command += " #{@settings['temp_dir']}track#{@track}_#{@trial}.wav" #output name
		unless @settings['verbose'] : command += " 2>&1" end # hide the output of cdparanoia output
#puts "command = #{command}"
		`#{command}` #Launch the cdparanoia command
	end
	
	def get_digest
		digest = Digest::MD5.hexdigest(File.read("#{@settings['temp_dir']}track#{@track}_1.wav"))
		@gui.append_2_log("MD5 Digest is #{digest}\n\n")
	end
end

class Encode
	def initialize(track, settings)
		@track = track
		@settings = settings
		@gui = @settings['log']
		@trackname = @settings['cd'].tracklist[@track - 1] 
		percentages()
		filenames()
		if @settings['flac'] : do_flac() end
		if @settings['vorbis'] : do_vorbis() end
		if @settings['mp3'] : do_mp3() end
		if @settings['wav'] : do_wav() end
		if @settings['other'] : do_other end
		File.delete("#{@settings['temp_dir']}track#{@track}_1.wav") #We don't need the wav file after it's encoded.
	end
	
	def percentages
		@codecs = 0
		if @settings['flac']: @codecs += 1 end
		if @settings['vorbis']: @codecs += 1 end
		if @settings['mp3']: @codecs += 1 end
		if @settings['wav']: @codecs += 1 end
		if @settings['other']: @codecs += 1 end
		@complete = @settings['enc_percentages'].shift()
	end
		
	def filenames #Changes the different %fields to usefull values
		@filename = '/' + @settings['destination_dir'].dup()
		@filename.gsub!('%n', sprintf("%02d", @track))
		@filename.gsub!('%t', @trackname)
		@dirnames = @settings['directories'].dup()
	end
		
	def do_flac
		filename = @dirnames.shift() + @filename + '.flac'
		if !@settings['flacsettings'] : @settings['flacsettings'] = '--best' end
		flac(filename)
		update_playlist('flac', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_vorbis
		filename = @dirnames.shift() + @filename + '.ogg'
		if !@settings['vorbissettings'] : @settings['vorbissettings'] = '-q 6' end
		vorbis(filename)
		update_playlist('vorbis', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs 
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_mp3
		@possible_lame_tags = ['A CAPPELLA', 'ACID', 'ACID JAZZ', 'ACID PUNK', 'ACOUSTIC', 'ALTERNATIVE', 'ALT. ROCK', 'AMBIENT', 'ANIME', 'AVANTGARDE', \
'BALLAD', 'BASS', 'BEAT', 'BEBOB', 'BIG BAND', 'BLACK METAL', 'BLUEGRASS', 'BLUES', 'BOOTY BASS', 'BRITPOP', 'CABARET', 'CELTIC', 'CHAMBER MUSIC', 'CHANSON', \
'CHORUS', 'CHRISTIAN GANGSTA RAP', 'CHRISTIAN RAP', 'CHRISTIAN ROCK', 'CLASSICAL', 'CLASSIC ROCK', 'CLUB', 'CLUB-HOUSE', 'COMEDY', 'CONTEMPORARY CHRISTIAN', \
'COUNTRY', 'CROSSOVER', 'CULT', 'DANCE', 'DANCE HALL', 'DARKWAVE', 'DEATH METAL', 'DISCO', 'DREAM', 'DRUM & BASS', 'DRUM SOLO', 'DUET', 'EASY LISTENING', \
'ELECTRONIC', 'ETHNIC', 'EURODANCE', 'EURO-HOUSE', 'EURO-TECHNO', 'FAST-FUSION', 'FOLK', 'FOLKLORE', 'FOLK/ROCK', 'FREESTYLE', 'FUNK', 'FUSION', 'GAME', \
'GANGSTA RAP', 'GOA', 'GOSPEL', 'GOTHIC', 'GOTHIC ROCK', 'GRUNGE', 'HARDCORE', 'HARD ROCK', 'HEAVY METAL', 'HIP-HOP', 'HOUSE', 'HUMOUR', 'INDIE', 'INDUSTRIAL', \
'INSTRUMENTAL', 'INSTRUMENTAL POP', 'INSTRUMENTAL ROCK', 'JAZZ', 'JAZZ+FUNK', 'JPOP', 'JUNGLE', 'LATIN', 'LO-FI', 'MEDITATIVE', 'MERENGUE', 'METAL', 'MUSICAL', \
'NATIONAL FOLK', 'NATIVE AMERICAN', 'NEGERPUNK', 'NEW AGE', 'NEW WAVE', 'NOISE', 'OLDIES', 'OPERA', 'OTHER', 'POLKA', 'POLSK PUNK', 'POP', 'POP-FOLK', 'POP/FUNK', \
'PORN GROOVE', 'POWER BALLAD', 'PRANKS', 'PRIMUS', 'PROGRESSIVE ROCK', 'PSYCHEDELIC', 'PSYCHEDELIC ROCK', 'PUNK', 'PUNK ROCK', 'RAP', 'RAVE', 'R&B', 'REGGAE', \
'RETRO', 'REVIVAL', 'RHYTHMIC SOUL', 'ROCK', 'ROCK & ROLL', 'SALSA', 'SAMBA', 'SATIRE', 'SHOWTUNES', 'SKA', 'SLOW JAM', 'SLOW ROCK', 'SONATA', 'SOUL', 'SOUND CLIP', \
'SOUNDTRACK', 'SOUTHERN ROCK', 'SPACE', 'SPEECH', 'SWING', 'SYMPHONIC ROCK', 'SYMPHONY', 'SYNTHPOP', 'TANGO', 'TECHNO', 'TECHNO-INDUSTRIAL', 'TERROR', 'THRASH METAL', \
'TOP 40', 'TRAILER', 'TRANCE', 'TRIBAL', 'TRIP-HOP', 'VOCAL']
		filename = @dirnames.shift() + @filename + '.mp3'
		if !@settings['mp3settings'] : @settings['mp3settings'] = "--preset fast standard" end
		if @possible_lame_tags.include?(@settings['cd'].genre.upcase): genre = @settings['cd'].genre else genre = 'other' end
		mp3(filename, genre)
		update_playlist('mp3', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs 
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_wav	
		filename = @dirnames.shift() + @filename + '.wav'
		wav(filename)
		update_playlist('wav', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs 
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
	
	def do_other
		@othersettings = @settings['othersettings'].dup() # make a duplication, because these values change per track
		@othersettings.gsub!('%n', sprintf("%02d", @track)) # %n = tracknumber
		@othersettings.gsub!('%t', '"' + @trackname + '"') # %t = trackname
		@othersettings.gsub!('%o', '"' + (temp = @dirnames.shift() + @filename) + '"') # %o = output filename
 		@othersettings.gsub!('%i', "#{@settings['temp_dir']}track#{@track}_1.wav") # %i = input filename
 		puts "command = #{@settings['othersettings']}" if @settings['debug']
 		`#{@settings['othersettings']}` # Run the command
 		
 		extension = false #file extension
 		if @settings['othersettings'].rindex('.') : extension = @settings['othersettings'][@settings['othersettings'].rindex('.')..-1] end
 		update_playlist(extension, temp + extension) if (@settings['playlist'] && extension)
		increase = (@complete - @gui.encoding_progress) / @codecs
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
	
	def flac(filename)
		`flac #{@settings['flacsettings']} -o "#{filename}" --tag ARTIST="#{@settings['cd'].artist}" --tag TITLE="#{@trackname}" --tag ALBUM="#{@settings['cd'].album}" --tag DATE="#{@settings['cd'].year}" --tag TRACKNUMBER="#{@track}" --tag GENRE="#{@settings['cd'].genre}" #{@settings['temp_dir']}track#{@track}_1.wav#{" 2>&1" unless $verbose}`
	end
	
	def vorbis(filename)
		`oggenc -o "#{filename}" #{@settings['vorbissettings']} --date "#{@settings['cd'].year}" --tracknum "#{@track}" --title "#{@trackname}" --album "#{@settings['cd'].album}" --artist "#{@settings['cd'].artist}" --genre "#{@settings['cd'].genre}" #{@settings['temp_dir']}track#{@track}_1.wav#{" 2>&1" unless $verbose}`
	end
	
	def mp3(filename, genre)
		`lame --id3v2-only #{@settings['mp3settings']} --tt "#{@trackname}" --ta "#{@settings['cd'].artist}" --tl "#{@settings['cd'].album}" --ty "#{@settings['cd'].year}" --tn "#{@track}" --tg "#{genre}" #{@settings['temp_dir']}track#{@track}_1.wav "#{filename}"#{" 2>&1" unless $verbose}`
	end
	
	def wav(filename)
		require 'fileutils'
		FileUtils.cp("#{@settings['temp_dir']}track#{@track}_1.wav", filename)
	end
	
	def update_playlist(format, filename)
		m3ufile = File.new(File.dirname(filename) + "/#{@settings['cd'].artist} - #{@settings['cd'].album} (#{format}).m3u", 'a')
		m3ufile.puts "#{File.basename(filename)}"
		m3ufile.close()
	end
end

class Rubyripper
	def initialize(settings)
		@settings = settings
		if check_configuration && test_for_deps && prepare_dirs # tests for correctness settings and the presents of codecs and rippers
			update_gui() # Give some info about the cdrom-player, the codecs, the ripper, cddb_info
			compute_percentage() # Do some pre-work to get the progress updater working later on
			temp_dir() # Check if the temporary dir exists, if not: make it if basedir does exist or fallback to /tmp/rr/.
			require 'digest/md5' # Needed for secure class, only have to load them ones here.
			ripper() # Function to control Secure_rip and Encode class
			cleanup() # Delete any temporary files
		end
	end
	
	def check_configuration
		if @settings['tracks_to_rip'].size == 0
			@settings['instance'].update("error", "Please select at least one track.")
		end
		
		temp = Cddb.new(@settings['cdrom'], @settings['instance'])
		if @settings['cd'].cd_playtime != temp.cd_playtime # Test if the user has put in another CD to fool me ;)
			@settings['instance'].update("error", "The Gui doesn't match inserted cd.\nPlease press Reload CD first.")
 			return false
		end
		
		unless @settings['flac'] || @settings['vorbis'] || @settings['mp3'] || @settings['wav'] || @settings['other'] # Test if at least one codec is selected
			@settings['instance'].update("error", "No codecs are selected!\nRip is canceled.")
			return false
 		end
 		
		#Make sure errors get at least as much corrected as non-errors.
 		if @settings['req_matches_all'] > @settings['req_matches_errors'] : @settings['req_matches_errors'] = @settings['req_matches_all'] end

		return true
	end
	
	def test_for_deps()
		{"ripper" => "cdparanoia", "flac" => "flac", "vorbis" => "oggenc", "mp3" => "lame"}.each do |setting, binary|
			if @settings[setting] && !File.exists?(`which #{binary}`.strip())
				@settings['instance'].update("error", "#{binary.capitalize} was not found on your system.\nRip is canceled.")
				return false
			end
		end
		return true
	end
	
	def prepare_dirs # Look it the dirs already exists, create them, prepare the filename for the Encode class while we're at it
		if @settings['destination_dir'].include?("/%b") && @settings['cd'].album[0,1] == '.' #If a dir or file starts with an album and the album starts with a dot 
			@settings['cd'].album.sub!(/\.*/, "_") #remove those dot(s) from the albumname and replace them with a single underscore, otherwise they're hidden files in linux.
		end
		
		{'%a' => @settings['cd'].artist, '%b' => @settings['cd'].album, '%g' => @settings['cd'].genre, '%y' => @settings['cd'].year}.each do |key, value|
			@settings['destination_dir'].gsub!(key, value) # change the keys for the values that stay the same for the whole cd
			@settings['othersettings'].gsub!(key, '"' + value + '"') # do the same for the external codec command
		end

		@settings['destination_dir'] = File.expand_path(@settings['destination_dir']) # Replace any ~/ with the relevant information, such as /home/username/
		
		@settings['directories'] = Array.new() # Later on the log will be saved in these directories.
		['flac', 'vorbis', 'mp3', 'wav', 'other'].each do |codec|
			if @settings[codec] : @settings['directories'] << File.dirname(@settings['destination_dir']).sub("%f", codec) end
		end
		
		require 'ftools' # Adding some extra File functions like makedirs
		
		@settings['directories'].each do |dir| # Test if directory already exists
			if File.directory?(dir)
				@settings['instance'].update("error", "Directory #{dir} already exists.\nRip is canceled.")
				return false #Stop the rip
			else 
				File.makedirs(dir) #make the necessary (sub)dir(s)
			end
		end
		
		@settings['destination_dir'] = File.basename(@settings['destination_dir']) # The dirs are already made and saved in @settings['directories']
		@gui = @settings['log'] = Gui_support.new(@settings['instance'], @settings['directories']) # The gui class :)
		return true
	end
	
	def update_gui
		@gui.append_2_log("Cdrom player used to rip: #{@settings['cd'].devicename}\n")
		@gui.append_2_log("Cdrom offset used: #{@settings['offset']}\n\n")
		@gui.append_2_log("Ripper used: cdparanoia (#{if @settings['rippersettings']: @settings['rippersettings'] else 'default settings' end})\n")
		@gui.append_2_log("Matches required for all chunks: #{@settings['req_matches_all']}\n")
		@gui.append_2_log("Matches required for erronous chunks: #{@settings['req_matches_errors']}\n\n")
		
		@gui.append_2_log("Codec(s) used:\n")
		if @settings['flac']: @gui.append_2_log("-flac \t-> #{@settings['flacsettings']} (#{`flac --version`.strip})\n") end
		if @settings['vorbis']: @gui.append_2_log("-vorbis\t-> #{@settings['vorbissettings']} (#{`oggenc --version`.strip})\n") end
		if @settings['mp3']: @gui.append_2_log("-mp3\t-> #{@settings['mp3settings']}\n(#{`lame --version`.split("\n")[0]})\n") end
		if @settings['wav']: @gui.append_2_log("-wav\n") end
		if @settings['other'] : @gui.append_2_log("-other\t-> #{@settings['othersettings']}\n") end
		@gui.append_2_log("\nCDDB INFO\n")
		@gui.append_2_log("\nArtist\t= " + @settings['cd'].artist)
		@gui.append_2_log("\nAlbum\t= " + @settings['cd'].album)
		@gui.append_2_log("\nYear\t= " + @settings['cd'].year)
		@gui.append_2_log("\nGenre\t= " + @settings['cd'].genre)
		@gui.append_2_log("\nTracks\t= " + @settings['cd'].audiotracks.to_s + "\n\n")
		@settings['cd'].audiotracks.times do |track|
			@gui.append_2_log("#{sprintf("%02d", track + 1)} - #{@settings['cd'].tracklist[track]}\n")
		end
		@gui.append_2_log("\nSTATUS\n\n")
	end
	
	def compute_percentage
		@settings['rip_percentages'] = Array.new() #cumulative progress
		total_sectors = 0.0 # It can be that the user doesn't want to rip all tracks, so calculate it
		@settings['tracks_to_rip'].each{|track| total_sectors += @settings['cd'].length_in_sectors[track - 1]}
		
		@settings['tracks_to_rip'].each do |track|
			if track == @settings['tracks_to_rip'][-1] # Ensure last value is 1.0
				@settings['rip_percentages'] << 1.0 
			elsif track == @settings['tracks_to_rip'][0] # Cannot refer to the previous value of rip_percentages, since it has none
				@settings['rip_percentages'] << @settings['cd'].length_in_sectors[track - 1] / total_sectors
			else # progress when track is done: length_track / total + previous amount
				@settings['rip_percentages'] << (@settings['cd'].length_in_sectors[track - 1] / total_sectors) + @settings['rip_percentages'][-1]
			end
 		end
 		@settings['enc_percentages'] = @settings['rip_percentages'].dup #Make a copy for the encoding progress.
 	end
 	
	def temp_dir
		if (File.directory?(File.dirname(@settings['temp_dir'])) || File.symlink?(File.dirname(@settings['temp_dir'])))
			if @settings['temp_dir'][-1,1] != '/': @settings['temp_dir'] += '/' end #Make sure the directory ends with a /
		else @settings['temp_dir'] = "/tmp/rr/"
			if @settings['verbose']: puts "#{@settings['temp_dir']} is used as temporary directory" end
		end
		unless File.directory?(@settings['temp_dir']) || File.symlink?(@settings['temp_dir']): Dir.mkdir(@settings['temp_dir']) end # If the directory not yet exists, create it
	end
		
	def ripper
		@last_track_ripped = false #necessary because of the threaded encoding
		@tracks_to_encode = Array.new() #necessary because of the threaded encoding
		@gui.update_ripping_progress(0.0, "ripper") # Give a hint to the gui that ripping has started
		@settings['tracks_to_rip'].each do |track|
			@secure_rip = Secure_rip.new(track, @settings)
			@gui.track_finished(track)
			@tracks_to_encode << track
			if @settings['threading']
				unless @encoding
					@encoding = Thread.new {encode(track)} #Let the encoding take place in the background.
					@encoding.priority = -1 #Encoding gets lower priority than ripping
				end
			else
				@encode_instance = Encode.new(track, @settings)
			end
		end
		
		@last_track_ripped = true
		if @settings['threading']: @encoding.join() end #Wait untill tracks are encoded, before cleaning up
		@gui.finished(@settings['req_matches_all'], @settings['req_matches_errors'], @settings['max_tries'], "ripper")
	end

	def encode(track)
		until @last_track_ripped && @tracks_to_encode.length == 0
			while track = @tracks_to_encode.shift
				@encode_instance = Encode.new(track, @settings)
			end
			if @tracks_to_encode.length == 0 && @last_track_ripped == false
				sleep(1)
			end
		end
	end
	
	def cleanup
		if File.directory?(@settings['temp_dir']) || File.symlink?(@settings['temp_dir'])
			Dir.foreach(@settings['temp_dir']) do |file|
				if file[0,1] != '.' && File.file?("#{@settings['temp_dir']}#{file}"): File.delete("#{@settings['temp_dir']}#{file}") end
			end
			unless File.symlink?(@settings['temp_dir']): Dir.delete(@settings['temp_dir']) end # Don't delete symlinks. They are there for a purpose
		end
	end
end
