#!/usr/bin/env ruby

if (ARGV.include?('--help') || ARGV.include?('-h'))
	puts "Rubyripper on commandline accepts the following parameters:\n
	-h --help \t\t Display this info
	-c --configure \t\t Configure default settings
	-a --all \t\t Rip all tracks. Skip any questions about track selection.
	"
	exit()
end

class Gui_CLI
	if File.exists?('rr_lib.rb')
		require './rr_lib.rb' # run from local directory
	else
		require 'rr_lib.rb' # installed in ruby directory
	end
	
	def initialize()
		@ripping_log = ""
		@ripping_progress = 0.0
		@encoding_progress = 0.0
		@settings = {"flac" => false, "flacsettings" => "--best", "vorbis" => false, "vorbissettings" => "-q6", "mp3" => false, "mp3settings" => "--preset fast standard", "wav" => false, "other" => false, "othersettings" => '', "playlist" => true, "cdrom" => "/dev/cdrom", "offset" => 0, "threading" => false, "rippersettings" => '', "max_tries" => 0, "destination_dir" => "~/%f/%a (%y) %b/%n - %t", "verbose" => false, "debug" => false, "temp_dir" => "/tmp/rr/", "instance" => self, "eject" => false, "req_matches_errors" => 2, "req_matches_all" => 2, "site" => "freedb.org", "username" => "anonymous", "hostname" => "my_secret.com", "first_hit" => false, "edit" => false, "freedb" => true} # some default settings
		if File.exist?(File.expand_path("~/.rubyripper_settings"))
			load_settings()
 			if ARGV.include?('-c') || ARGV.include?('--configure'): edit_settings() end
		else
			edit_settings() # Always set up settings if no config file exists.
		end
		
		main()
	end
	
	def show_settings # simple overview of current settings, with numbers so a user can choose what to change
		puts "#### RUBYRIPPER SETTINGS ####"
		puts ""
		puts "  ## ENCODING ##  "
 		puts "1)  flac [#{@settings['flac'] ? '*' : ' '}] , settings: #{@settings['flacsettings']}"
 		puts "2)  vorbis [#{@settings['vorbis'] ? '*' : ' '}] , settings: #{@settings['vorbissettings']}"
 		puts "3)  mp3 [#{@settings['mp3'] ? '*' : ' '}] , settings: #{@settings['mp3settings']}"
 		puts "4)  wav [#{@settings['wav'] ? '*' : ' '}]"
 		puts "5)  other [#{@settings['other'] ? '*' : ' '}] , settings: #{@settings['othersettings']}"
 		puts "6)  playlist support [#{@settings['playlist'] ? '*' : ' '}]"
 		puts "7)  encode while ripping [#{@settings['threading'] ? '*' : ' '}]"
 		puts ""
 		puts "  ## RIPPING ##  "
 		puts "8)  cdrom : #{@settings['cdrom']} with offset #{@settings['offset']}"
 		puts "9)  passing extra cdparanoia parameters : #{@settings['rippersettings'] != '' ? @settings['rippersettings'] : 'none'}"
 		puts "10) match all chunks : #{@settings['req_matches_all']}\tmatch erronous chunks : #{@settings['req_matches_errors']}"
 		puts "11) maximum trials : #{@settings['max_tries'] == 0 ? "no maximum" : @settings['max_tries']}"
 		puts "12) eject disc after ripping [#{@settings['eject'] ? '*' : ' '}]"
 		puts "13) temporary files directory : #{@settings['temp_dir']}"
 		puts "14) destination directory : #{@settings['destination_dir']}"
 		puts ""
 		puts "  ## FREEDB ##  "
 		puts "15) fetch cd info with freedb [#{@settings['freedb'] ? '*' : ' '}]"
 		puts "16) always use first hit [#{@settings['first_hit'] ? '*' : ' '}]"
 		puts "17) site = #{@settings['site']}"
		puts "18) username = #{@settings['username']}"
 		puts "19) hostname = #{@settings['hostname']}"
 		puts "20) edit freedb data before ripping [#{@settings['edit'] ? '*' : ' '}]"
 		puts ""
 		puts "99) Don't change any setting"
 		puts ""
	end

	def edit_settings
		show_settings()
		while (number = get_answer("Please type the number of the setting you wish to change : ", "number")) != 99
			if number == 1 : set_codec('flac')
			elsif number == 2 : set_codec('vorbis')
			elsif number == 3 : set_codec('mp3')
			elsif number == 4 : if @settings['wav'] : puts "wav disabled" ; @settings['wav'] = false else puts "wav enabled" ; @settings['wav'] = true end
			elsif number == 5 : set_codec('other')
			elsif number == 6 : if @settings['playlist'] : puts "playlist disabled" ; @settings['playlist'] = false else puts "playlist enabled" ; @settings['playlist'] = true end
			elsif number == 7 : if @settings['threading'] : puts "threading disabled" ; @settings['threading'] = false else puts "threading enabled" ; @settings['threading'] = true end
			elsif number == 8 : set_cdrom()
			elsif number == 9 : @settings['rippersettings'] = get_answer("Which options to pass to cdparanoia? : ", "open")
			elsif number == 10 : @settings['req_matches_all'] = get_answer("How many times must all chunks be matched? : ", "number")
			elsif number == 11 : @settings['req_matches_errors'] = get_answer("How many times must erronous chunks be matched? : ", "number")
			elsif number == 12 : if @settings['eject'] : puts "eject disabled" ; @settings['eject'] = false else puts "eject enabled" ; @settings['eject'] = true end
			elsif number == 13 : @settings['temp_dir'] = get_answer("Temporary directory : ", "open")
			elsif number == 14 : @settings['destination_dir'] = get_answer("(%a = artist, %b = album, %g = genre, %y = year, %f = codec, %t = trackname, %n = tracknumber) \nDestination dir : ", "open")
			elsif number == 15 : if @settings['freedb'] : puts "freedb disabled" ; @settings['freedb'] = false else puts "freedb enabled" ; @settings['freedb'] = true end
			elsif number == 16 : if @settings['first_hit'] : puts "first hit disabled" ; @settings['first_hit'] = false else puts "first hit enabled" ; @settings['first_hit'] = true end
			elsif number == 17 : @settings['site'] = get_answer("Freedb mirror: ", "open")
			elsif number == 18 : @settings['username'] = get_answer("Username : ", "open")
			elsif number == 19 : @settings['hostname'] = get_answer("Hostname : ", "open")
			elsif number == 20 : if @settings['edit'] : puts "freedb editing disabled" ; @settings['edit'] = false else puts "freedb editing enabled" ; @settings['edit'] = true end 
			else puts "Number #{number} is not an option!\nPlease try again."
			end
			puts ""
			show_settings()
		end
		save_settings()
		unless get_answer("Do you want to continue with the program now? (y/n) : ", "yes") : exit() end
	end
	
	def set_codec(codec)
		@settings[codec] = get_answer("Do you want to enable #{codec} encoding? (y/n) : ", "yes")
		if @settings[codec]
			if get_answer("Do you want to change the encoding parameters for #{codec}? (y/n) : ", "yes")
				if codec == 'other' : puts "%a = artist, %b = album, %g = genre, %y = year, %t = trackname, %n = tracknumber, %i = inputfile, %o = outputfile (don't forget extension)" end
				@settings[codec + 'settings'] = get_answer("Encoding paramaters for #{codec} : ", "open")
			end
		end
	end
	
	def set_cdrom
		@settings['cdrom'] = get_answer("Cdrom device : ", "open")
		@settings['offset'] = get_answer("Offset for drive : ", "number")
	end
	
	def save_settings #save to config file
		file = File.new(File.expand_path("~/.rubyripper_settings"), 'w')
#ENCODING
		file.puts "flac=#{@settings['flac']}"
		file.puts "vorbis=#{@settings['vorbis']}"
		file.puts "mp3=#{@settings['mp3']}"
		file.puts "wav=#{@settings['wav']}"
		file.puts "other=#{@settings['other']}"
		file.puts "flacsettings=#{@settings['flacsettings']}"
		file.puts "vorbissettings=#{@settings['vorbissettings']}"
		file.puts "mp3settings=#{@settings['mp3settings']}"
		file.puts "othersettings=#{@settings['othersettings']}"
		file.puts "playlist=#{@settings['playlist']}"
		file.puts "threading=#{@settings['threading']}"
#RIPPING
		file.puts "cdrom=#{@settings['cdrom']}"
		file.puts "offset=#{@settings['offset']}"
		file.puts "rippersettings=#{@settings['rippersettings']}"
		file.puts "req_matches_all=#{@settings['req_matches_all']}"
		file.puts "req_matches_errors=#{@settings['req_matches_errors']}"
		file.puts "max_tries=#{@settings['max_tries']}"
		file.puts "eject=#{@settings['eject']}"
		file.puts "temp_dir=#{@settings['temp_dir']}"
		file.puts "destination_dir=#{@settings['destination_dir']}"
#FREEDB
		file.puts "freedb=#{@settings['freedb']}"
		file.puts "first_hit=#{@settings['first_hit']}"
		file.puts "site=#{@settings['site']}"
		file.puts "username=#{@settings['username']}"
		file.puts "hostname=#{@settings['hostname']}"
		file.puts "edit=#{@settings['edit']}"
		file.close()
	end
	
	def load_settings # load config file
		file = File.new(File.expand_path("~/.rubyripper_settings"),'r')
		while line = file.gets
#ENCODING
			if line =~ /flac=true/ : @settings['flac'] = true
			elsif line =~ /vorbis=true/ : @settings['vorbis'] = true
			elsif line =~ /mp3=true/ : @settings['mp3'] = true
			elsif line =~ /wav=true/ : @settings['wav'] = true
			elsif line =~ /other=true/ : @settings['other'] = true
			elsif line =~ /flacsettings=/ : @settings['flacsettings'] = $'.rstrip
			elsif line =~ /vorbissettings=/ : @settings['vorbissettings'] = $'.rstrip
			elsif line =~ /mp3settings=/ : @settings['mp3settings'] = $'.rstrip
			elsif line =~ /othersettings=/ : @settings['othersettings'] = $'.rstrip
			elsif line =~ /playlist=false/ : @settings['playlist'] = false
			elsif line =~ /threading=true/ : @settings['threading'] = true
#RIPPING
			elsif line =~ /cdrom=/ : @settings['cdrom'] = $'.rstrip
			elsif line =~ /offset=/ : @settings['offset'] = $'.rstrip.to_i
			elsif line =~ /rippersettings=/ : @settings['rippersettings'] = $'.rstrip
			elsif line =~ /req_matches_all=/ : @settings['req_matches_all'] = $'.rstrip.to_i
			elsif line =~ /req_matches_errors=/ : @settings['req_matches_errors'] = $'.rstrip.to_i
			elsif line =~ /max_tries=/ : @settings['max_tries'] = $'.rstrip.to_i
			elsif line =~ /eject=true/ : @settings['eject'] = true
			elsif line =~ /temp_dir=/ : @settings['temp_dir'] = $'.rstrip
			elsif line =~ /destination_dir=/ : @settings['destination_dir'] = $'.rstrip
#FREEDB
			elsif line =~ /freedb=false/ : @settings['freedb'] = false
			elsif line =~ /first_hit=false/ : @settings['first_hit'] = false
			elsif line =~ /site=/ : @settings['site'] = $'.rstrip
			elsif line =~ /username=/ : @settings['username'] = $'.strip
			elsif line =~ /hostname=/ : @settings['hostname'] = $'.strip
			elsif line =~ /edit=true/ : @settings['edit'] = true
			end
		end
		file.close()
		if @settings['rippersettings'] == "''" : @settings['rippersettings'] = '' end
		if @settings['flacsettings'] == "''" : @settings['flacsettings'] = '' end
		if @settings['vorbissettings'] == "''" : @settings['vorbissettings'] = '' end
		if @settings['mp3settings'] == "''" : @settings['mp3settings'] = '' end
		if @settings['othersettings'] == "''" : @settings['othersettings'] = '' end
	end
	
	def show_freedb
		puts ""
		puts "FREEDB INFO"
		puts "1) Artist : #{@settings['cd'].artist}"
		puts "2) Album : #{@settings['cd'].album}"
		puts "3) Genre : #{@settings['cd'].genre}"
		puts "4) Year : #{@settings['cd'].year}"
		@settings['cd'].audiotracks.times{|index| puts "#{index + 5}) Track #{index + 1} : #{@settings['cd'].tracklist[index]}"}
		puts ""
	end
	
	def edit_freedb
		show_freedb()
		puts "99) Don't change any setting"
 		puts ""
		while (number = get_answer("Please type the number of the setting you wish to change :", "number")) != 99
			if number == 1 : @settings['cd'].artist = get_answer("Artist : ", "open")
			elsif number == 2 : @settings['cd'].album = get_answer("Album : ", "open")
			elsif number == 3 : @settings['cd'].genre = get_answer("Genre : ", "open")
			elsif number == 4 : @settings['cd'].year = get_answer("Year : ", "open")
			else
				if number <= @settings['cd'].audiotracks + 4 #Check if tracknumber is in valid range
					@settings['cd'].tracklist[number - 5] = get_answer("Track #{number - 4} :", "open")
				else
					puts "This is not a valid number."
				end
			end	
		end
	end
	
	def get_answer(question, answer)
		succes = false
		while !succes 
			if answer == "yes"
				answers = ["yes", "y", "no", "n"]
				STDOUT.print(question)
				input = STDIN.gets.strip
				if answers.include?(input)
					if input == 'y' || input == "yes" : return true else return false end
				else puts "Please answer yes or no" end
			elsif answer == "open"
				STDOUT.print(question)
				return STDIN.gets.strip
			elsif answer == "number"
				STDOUT.print(question)
				input = STDIN.gets.strip.to_i
				if input.size != 0 : return input
				else puts "#{input} is not a valid number! Please enter a valid number" end
			else puts "We should never get here" ; puts "answer = #{answer}, question = #{question} (error)"
 			end
		end
	end
	
	def cdrom
		@settings['cd'] = Cddb.new(@settings['cdrom'], @settings['instance']) # Analyze audio-cd, don't look at cddb yet.
		if @settings['cd'].audiotracks != 0
			puts "Audio-disc found, number of tracks : #{@settings['cd'].audiotracks}, total playlength : #{@settings['cd'].cd_playtime}"
			if @settings['freedb'] : puts "Fetching freedb info..." ; @settings['cd'].freedb(@settings, !@settings['first_hit']) end
			if @settings['edit'] : edit_freedb() else show_freedb() end
		else 
			puts "No audio-disc found, exiting..."; exit()
		end
	end
	
	def choose_freedb(choices)
		puts "Freedb reported multiple possibilities."
		choices.each_index{|index| puts "#{index + 1}) #{choices[index]}"}
		choice = get_answer("Please type the number of the one you prefer? : ", "number")
		@settings['cd'].freedb_choice(choice - 1)
	end
	
	def main
		cdrom() # Fetch CD info
		tracklist() # Which tracks should be ripped?
		@rubyripper = Rubyripper.new(@settings)
	end

  	def update(modus, value)
		if modus == "ripping_progress"
			progress = "%.3g" % (value * 100)
			puts "Ripping progress (#{progress} %)"
		elsif modus == "encoding_progress"
			progress = "%.3g" % (@encoding_progress * 100)
			puts "Encoding progress (#{progress} %)"
		elsif modus == "log_change"
			print value
		elsif modus == "error"
			print value
			print "\n"
			if get_answer("Do you want to change your settings? (y/n) : ", "yes") : edit_settings() end
		elsif modus == "cddb_hit" && value != false
			choose_freedb(value)
		end
	end

	def tracklist # Fill @settings['tracks_to_rip']
		@settings['tracks_to_rip'] = Array.new
		@settings['cd'].audiotracks.times{|number| @settings['tracks_to_rip'] << number + 1} # Start with all tracks selected
		if ARGV.include?('-a') || ARGV.include?('--all') || get_answer("\nShould all tracks be ripped ? (y/n) ", "yes")
			puts "Tracks to rip are #{@settings['tracks_to_rip'].join(", ")}"
		else
			succes = false
			while !succes
				puts "Current selection of tracks : #{@settings['tracks_to_rip'].join(', ')}"
				number = get_answer("Enter 1 for entering the tracknumbers you want to remove.\nEnter 2 if for entering the tracks you want to keep.\nYour choice: ", "number")
				if number == 1
					print "Type the numbers of the tracks you want to remove and separate them with a space: "
					answer = STDIN.gets.strip.split
					anwer.each_index{|index| answer[index] = answer[index].to_i} #convert to integers
					@settings['tracks_to_rip'] -= answer # don't you just love ruby math? [1,2,3,4,5] - [3,4] = [1,2,5]
				elsif number == 2
					print "Type the numbers of the tracks you want to keep and separate them with a space: "
					answer = STDIN.gets.strip.split
					answer.each_index{|index| answer[index] = answer[index].to_i} #convert to integers
					@settings['tracks_to_rip'].each{|track| unless answer.include?(track) : @settings['tracks_to_rip'].delete(track) end}
				else
					puts "#{number} is not a valid number! Please enter 1 or 2!\n"
				end
				if number == 1 || number == 2
					puts "Current selection of tracks : #{@settings['tracks_to_rip'].join(', ')}"
					succes = !get_answer("Do you want to make any changes? (y/n) : ", "yes")
				end
			end
		end
	end
end	

Gui_CLI.new()
