#!/usr/bin/env ruby

begin
	require 'libglade2'
rescue LoadError
	puts "The ruby-libglade2 library could not be found. Is it installed?"; exit()
end

if File.exists?('rr_lib.rb')
	require './rr_lib.rb' # run from local directory
else
	require 'rr_lib.rb' # installed in ruby directory
end

class Gui_GTK
	def initialize()
		@just_started = true # needed to speed up the arrival of the GUI
		@rubyripper = false
		@glade = GladeXML.new('rubyripper.glade') {|handler| method(handler)}
		
		@verbose_mode = @glade.get_widget('verbose_mode') ; @debug_mode = @glade.get_widget('debug_mode') ; @temp_dir = @glade.get_widget('tempdir')
		
		@flac_button = @glade.get_widget('flacbutton') ; @vorbis_button = @glade.get_widget('vorbisbutton') ; @mp3_button = @glade.get_widget('mp3button') ; @wav_button = @glade.get_widget('wavbutton') ; @other_button = @glade.get_widget('other_button')
		@threading = @glade.get_widget('threading') ; @flac_entry = @glade.get_widget('flac_entry'); @vorbis_entry = @glade.get_widget('vorbis_entry'); @mp3_entry = @glade.get_widget('mp3_entry'); @other_entry = @glade.get_widget('other_entry') ; @playlist_button = @glade.get_widget('playlist')

		@cdrom_entry = @glade.get_widget('cdrom') ; @ripper_entry = @glade.get_widget('ripper_settings') ; @offset_button = @glade.get_widget('offset_button') ; @destination_dir = @glade.get_widget('destination_dir') ; @req_matches_all = @glade.get_widget('req_matches_all') ; @req_matches_errors = @glade.get_widget('req_matches_errors') ; @maximum_tries = @glade.get_widget('maximum_tries') ; @eject = @glade.get_widget('eject')

		@statusview = @glade.get_widget('statusview') ;	@statuswindow = @glade.get_widget('statuswindow') ; @statusbar = @glade.get_widget('statusbar') ; @ripping_progress = @glade.get_widget('ripping_progress'); @encoding_progress = @glade.get_widget('encoding_progress')
		
		@freedb = @glade.get_widget('freedb'); @hostname = @glade.get_widget('freedb_hostname') ; @username = @glade.get_widget('freedb_username') ; @site = @glade.get_widget('freedb_server') ; @first_freedbhit = @glade.get_widget('first_freedbhit')
		
		if File.exist?(File.expand_path("~/.rubyripper_settings")): loadsettings() end
		if @debug_mode.active?: Thread.abort_on_exception = true end
		@ripping_log = ""
		@settings = Hash.new
		@cddb = Thread.new {cdrom}
		@cddb.priority = -1 # Give resources to show the GUI first
		@busy = true # Later needed to block unwanted clicks on buttons a.k.a. make it idiot-proof
	end
	
	def cdrom
		@statusbar.pop(10)
		@statusbar.push(2, "Analyzing CD data...")
		if @just_started: @artist = @glade.get_widget('artist') ; @album = @glade.get_widget('album') ; @year = @glade.get_widget('year'); @genre = @glade.get_widget('genre'); @table = @glade.get_widget('table'); @dialog1 = @glade.get_widget('dialog1') ; @exit_dialog = @glade.get_widget('exit_dialog'); @cdrom_text = @glade.get_widget('cdrom1'); @refresh_dialog = @glade.get_widget('refresh_dialog'); @notebook = @glade.get_widget('notebook'); @freedb_dialog = @glade.get_widget('freedb_dialog'); @freedb_combobox = @glade.get_widget('freedb_combobox') ; @just_started = false end
		
		if @cdrom_entry.text != '' : @cdrom = @cdrom_entry.text
		elsif RUBY_PLATFORM =~ /linux/ : @cdrom_entry.text = @cdrom = '/dev/cdrom'
		elsif RUBY_PLATFORM =~ /darwin/ : @cdrom_entry.text = @cdrom = '/dev/disc1'
		elsif RUBY_PLATFORM =~ /bsd/ : @cdrom_entry.text = @cdrom = '/dev/cdrom'
		end
		
		@settings['cd'] = Cddb.new(@cdrom, self, @verbose_mode.active?) # Analyze audio-cd, don't look at cddb yet.
		
		if @settings['cd'].audiotracks != 0 : create_table() ; @statusbar.pop(2) # if audiotracks == 0, this means no cd is found
		else @statusbar.pop(2); @statusbar.push(2, "No disc found in @cdrom"); Thread.new{sleep(2); @statusbar.pop(2)} end
		
		if @freedb.active? # Fetch the freedb data if user wants to
			freedb_options = {"hostname" => @hostname.text, "username" => @username.text, "site" => @site.text}
			@settings['cd'].freedb(freedb_options, !@first_freedbhit.active?)
		end
		@statusbar.push(10, "Ready")
		@busy = false
	end
	
	def create_table()
		tracknumber = Gtk::CheckButton.new('Select all')
		tracknumber.active = true
		trackname = Gtk::Label.new('Tracknames')
		length = Gtk::Label.new('Length')
		@checkboxes = Array.new ; @tracknames = Array.new ; track_length = Array.new
		
		@settings['cd'].audiotracks.times do |number| # For eacht track create an entry with a checkbox, trackname and length
			@checkboxes << Gtk::CheckButton.new((number + 1).to_s)
			@checkboxes[number].active = true
			@tracknames << Gtk::Entry.new
			@tracknames[number].text = "Track #{number + 1}"
			@tracknames[number].width_request = 300
			track_length << Gtk::Label.new(@settings['cd'].tracks_playtime[number])
		end
		
		table_rows = @settings['cd'].audiotracks + 1
		@table.resize(table_rows, 3)
		table_rows.times do |number|
			if number == 0
				@table.attach(tracknumber, 0, 1, 0, 1, Gtk::SHRINK, Gtk::SHRINK, 0, 6)
				@table.attach(trackname, 1, 2, 0, 1, Gtk::EXPAND, Gtk::SHRINK, 0, 6)
				@table.attach(length, 2, 3, 0, 1, Gtk::SHRINK, Gtk::SHRINK, 0, 6)
			else
				@table.attach(@checkboxes[number - 1], 0, 1, number, number + 1, Gtk::SHRINK, Gtk::SHRINK)
				@table.attach(@tracknames[number - 1], 1, 2, number, number + 1, Gtk::EXPAND, Gtk::SHRINK)
				@table.attach(track_length[number -1], 2, 3, number, number + 1, Gtk::SHRINK, Gtk::SHRINK)
			end
		end
		tracknumber.signal_connect("toggled") {tracknumber.active? ? @checkboxes.each{|box| box.active = true} : @checkboxes.each{|box| box.active = false}} #Toggle on/off al tracks
		@table.show_all
	end
	
	def update_table # Update table for cddb info
		@artist.text = @settings['cd'].artist ; @album.text = @settings['cd'].album ; @year.text = @settings['cd'].year ; @genre.text = @settings['cd'].genre
		number = 0
		@tracknames.each do |button|
			button.text = @settings['cd'].tracklist[number]
			number += 1
		end
	end

	def refresh_cd
		if ! @busy
			@busy = true; @statusbar.pop(10)
			Thread.new do
				@ripping_progress.fraction = 0.0 ; @ripping_progress.text = "Ripping not yet started (0 %)" ; @encoding_progress.fraction = 0.0 ; @encoding_progress.text = "Encoding not yet started (0 %)" 
				@statusview.buffer.text = '' ; @artist.text = 'Unknown' ; @genre.text = 'Unknown' ; @album.text = 'Unknown' ; @year.text = '0' # Reset track info
				@table.children.each{|child| child.destroy} # Reset track info
				cdrom()
			end
		else
			@refresh_dialog.show_all()
		end
	end

	def on_cdrom_editing_done()
		if @cdrom_entry.text != @cdrom # only set new value if it is different from current value
			if File.exists?(@cdrom_entry.text)
				@cddb = refresh_cd()
			else
				update("error", "There is no such device available as #{@cdrom_entry.text}.\nPlease enter the correct cdrom device.\n")
				@cdrom = @cdrom_entry.text # Set it to the new device anyway
			end
		end
		return false	#Don't close application
	end

	def update_cdinfo()
		@settings['cd'].artist = @artist.text
		@settings['cd'].album = @album.text
		@settings['cd'].genre = @genre.text
		@settings['cd'].year = @year.text if @year.text.to_i != 0
		@settings['cd'].audiotracks.times do |index|
			@settings['cd'].tracklist[index] = @tracknames[index].text
		end
	end

	def on_ripcdbutton_clicked(widget)
		if ! @busy
			@busy = true; @statusbar.push(8, "Busy ripping / encoding cd #{@settings['cd'].artist} - #{@settings['cd'].album}")
			@rip = Thread.new do
				if @debug_mode.active?: Thread.abort_on_exception = true end
				update_cdinfo()
				get_settings()
				@statuswindow.show_all()
				@ripping_progress.text = "Ripping started... (0 %)"
				@statusview.move_viewport(Gtk::SCROLL_ENDS, 8)
				@rubyripper = Rubyripper.new(@settings)
				@statusbar.pop(8) ; @busy = false ; @rubyripper = nil
			end
		end
	end

	def update(modus, value) #Threading here is not really necessary, because none of these are cpu intensive. Ruby can't pass return values outside a thread.
		if modus == "ripping_progress"
			progress = "%.3g" % (value * 100)
			@ripping_progress.text = "Ripping progress (#{progress} %)"
			@ripping_progress.fraction = value
		elsif modus == "encoding_progress"
			progress = "%.3g" % (value * 100)
			@encoding_progress.text = "Encoding progress (#{progress} %)"
			@encoding_progress.fraction = value
		elsif modus == "log_change"
			@statusview.buffer.insert(@statusview.buffer.end_iter, value) # First parameter is the last character + 1 in the log
			@statusview.scroll_to_iter(@statusview.buffer.end_iter, 0, false, 0, 0)
		elsif modus == "track_finished"
			@checkboxes[value - 1].active = false
		elsif modus == "finished"
			@busy = false
			@statusbar.pop(8)
		elsif modus == "error"
			@cdrom_text.text = value
			@dialog1.show_all()
		elsif modus == "cddb_hit"
			if value != false
				value.each{|freedb_hit| @freedb_combobox.append_text(freedb_hit)}
				@freedb_dialog.show_all()
				@statusbar.push(2, "Fetching freedb info...")
			else
				update_table() ; @statusbar.pop(2) # if value == false, cd_info is finished
			end
		else
			puts "Ehh.. There shouldn't be anything else. WTF?"
			puts "Secret modus = #{modus}"
		end
	end
	
#Now let's build the big hash with all settings in it, so Rubyripper can get started
	def get_settings()
		@settings['instance'] = self
		@settings['verbose'] = @verbose_mode.active? ? true : false
		@settings['debug'] = @debug_mode.active? ? true : false
		@settings['temp_dir'] = @temp_dir.text

		@settings['flac'] = @flac_button.active? ? true : false
		@settings['vorbis'] = @vorbis_button.active? ? true : false
		@settings['mp3'] = @mp3_button.active? ? true : false
		@settings['wav'] = @wav_button.active? ? true : false
		@settings['other'] = @other_button.active? ? true : false
				
		@settings['flacsettings'] = @flac_entry.text
		@settings['vorbissettings'] = @vorbis_entry.text
		@settings['mp3settings'] = @mp3_entry.text
		@settings['othersettings'] = @other_entry.text
		
		@settings['playlist'] = @playlist_button.active? ? true : false
		@settings['threading'] = @threading.active? ? true : false
		@settings['cdrom'] = @cdrom
		@settings['rippersettings'] = @ripper_entry.text
		@settings['offset'] = @offset_button.value.to_i
		@settings['destination_dir'] = @destination_dir.text
		@settings['max_tries'] = @maximum_tries.value.to_i
		@settings['eject'] = @eject.active? ? true : false
		@settings['req_matches_all'] = @req_matches_all.value.to_i
		@settings['req_matches_errors'] = @req_matches_errors.value.to_i
		tracklist()
	end

	def tracklist
		@settings['tracks_to_rip'] = Array.new
		@checkboxes.each do |box|
			if box.active?
				@settings['tracks_to_rip'] << box.label.to_i
			end
		end
	end

	def on_savebutton_clicked(widget)
		settings = File.new(File.expand_path("~/.rubyripper_settings"), 'w')
		if @flac_button.active?: settings.puts "flac=true" end
		if @mp3_button.active?: settings.puts "mp3=true" end
		if @vorbis_button.active?: settings.puts "vorbis=true" end
		if @wav_button.active?: settings.puts "wav=true" end
		if @other_button.active? : settings.puts "other=true" end
		settings.puts "flacsettings=#{@flac_entry.text}"
		settings.puts "mp3settings=#{@mp3_entry.text}"
		settings.puts "vorbissettings=#{@vorbis_entry.text}"
		settings.puts "rippersettings=#{@ripper_entry.text}"
		settings.puts "othersettings=#{@other_entry.text}"
		if @freedb.active?: settings.puts "freedb=true" end
		if @first_freedbhit.active? : settings.puts "first_hit=true" end
		if @eject.active? : settings.puts "eject=true" end
		settings.puts "hostname=#{@hostname.text}"
		settings.puts "username=#{@username.text}"
		settings.puts "site=#{@site.text}"
		settings.puts "cdrom=#{@cdrom_entry.text}"
		settings.puts "offset=#{@offset_button.value.to_i}"
		settings.puts "destination_dir=#{@destination_dir.text}"
		settings.puts "max_tries=#{@maximum_tries.value}"
		settings.puts "req_matches_all=#{@req_matches_all.value}"
		settings.puts "req_matches_errors=#{@req_matches_errors.value}"
		settings.puts "temp_dir=#{@temp_dir.text}"
		if @verbose_mode.active?: settings.puts "verbose=true" end
		if @threading.active?: settings.puts "threading=true" end
		if @playlist_button.active?: settings.puts "playlist=true" end
		if @debug_mode.active?: settings.puts "debug=true" end
		settings.close()
		@statusbar.push(1, "Settings saved in #{File.expand_path("~/.rubyripper_settings")}")
		Thread.new {sleep(2); @statusbar.pop(1)}
	end
		
	def loadsettings
		File.open(File.expand_path("~/.rubyripper_settings")) do |file|
			while line = file.gets
				if line =~ /flac=true/: @flac_button.active = true
				elsif line =~ /mp3=true/: @mp3_button.active = true
				elsif line =~ /vorbis=true/: @vorbis_button.active = true
				elsif line =~ /wav=true/: @wav_button.active = true
				elsif line =~ /other=true/: @other_button.active = true
				elsif line =~ /flacsettings=/: @flac_entry.text = $'.rstrip
				elsif line =~ /mp3settings=/: @mp3_entry.text = $'.rstrip
				elsif line =~ /vorbissettings=/: @vorbis_entry.text = $'.rstrip
				elsif line =~ /rippersettings=/: @ripper_entry.text = $'.rstrip
				elsif line =~ /othersettings=/ : @other_entry.text = $'.rstrip
				elsif line =~ /freedb=true/: @freedb.active = true
				elsif line =~ /first_hit=true/: @first_freedbhit.active = true
				elsif line =~ /eject=true/: @eject.active = true
				elsif line =~ /hostname=/: @hostname.text = $'.rstrip
				elsif line =~ /username=/: @username.text = $'.rstrip
				elsif line =~ /site=/: @site.text = $'.rstrip
				elsif line =~ /cdrom=/: @cdrom_entry.text = $'.rstrip
				elsif line =~ /offset=/: @offset_button.value = $'.rstrip.to_f
				elsif line =~ /destination_dir=/: @destination_dir.text = $'.rstrip
				elsif line =~ /max_tries=/: @maximum_tries.value = $'.rstrip.to_f
				elsif line =~ /req_matches_all=/: @req_matches_all.value = $'.rstrip.to_f
				elsif line =~ /req_matches_errors=/: @req_matches_errors.value = $'.rstrip.to_f
				elsif line =~ /temp_dir=/: @temp_dir.text = $'.rstrip
				elsif line =~ /verbose=true/: @verbose_mode.active = true
				elsif line =~ /threading=true/: @threading.active = true
				elsif line =~ /playlist=true/: @playlist_button.active = true
				elsif line =~ /debug=true/: @debug_mode.active = true
				end
			end
		end
		if @flac_entry.text == "''" : @flac_entry.text = '' end
		if @vorbis_entry.text == "''" : @vorbis_entry.text = '' end
		if @mp3_entry.text == "''" : @mp3_entry.text == '' end
		if @ripper_entry.text == "''" : @ripper_entry.text == '' end
		if @other_entry.text == "''" : @other_entry.text == '' end
	end
	
	def freedb_choice
		Thread.new do # Don't block the gui
			choice = @freedb_combobox.active
			if choice == -1 # no option is selected
				@settings['cd'].freedb_choice(0)
			else 
				@settings['cd'].freedb_choice(choice)
			end
		end
	end
	
	def on_reload_cd_clicked(widget)
		refresh_cd()
	end
	
	def on_reload_anyway_clicked(widget)
		end_all_processes()
		@busy = false
 		@rubyripper = false ; @cddb = refresh_cd()
	end
	
	def on_exit_anyway_clicked(widget)
		end_all_processes()
		clean_window()
	end
	
	def clean_window
		@statuswindow.hide_all # First hide it, then reset it's components
		@statusview.buffer.text = ''
		@ripping_progress.fraction = 0.0
		@ripping_progress.text = "Ripping not yet started (0 %)"
		@encoding_progress.fraction = 0.0
		@encoding_progress.text = "Encoding not yet started (0 %)" 
	end
	
	def end_all_processes
		if @busy
			if @rip: @rip.kill end
			`killall cdparanoia 2>&1`
			@statusbar.pop(8) ; @busy = false ; @rubyripper = nil
 		end
	end

	def on_exitbutton_clicked(widget, debug = false)
		if @busy: @exit_dialog.show_all else clean_window() end # First test if the program is idle
	end

	def on_window1_delete_event
		end_all_processes()
		Gtk.main_quit
	end
end

# Main program
if __FILE__ == $0
	Gtk.init
	Gui_GTK.new()
	Gtk.main
end
